# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_dist_utils.ipynb.

# %% auto 0
__all__ = ['setup_distributed', 'get_rank', 'primary', 'synchronize', 'get_world_size', 'reduce_sum', 'gather_grad', 'all_gather',
           'all_gatherv', 'all_reduce', 'rank0_to_all', 'reduce_loss_dict']

# %% ../nbs/07_dist_utils.ipynb 3
import torch
from torch import distributed as dist
import os
def setup_distributed():
    local_rank = int(os.environ['LOCAL_RANK']) if 'LOCAL_RANK' in os.environ else 0
    n_gpu = int(os.environ['WORLD_SIZE']) if 'WORLD_SIZE' in os.environ else 1
    is_distributed = n_gpu > 1
    if is_distributed:
        torch.cuda.set_device(local_rank)
        dist.init_process_group(backend="nccl", init_method="env://")
        synchronize()
    return is_distributed


def get_rank():
    if not dist.is_available():
        return 0

    if not dist.is_initialized():
        return 0

    return dist.get_rank()


def primary():
    if not dist.is_available():
        return True

    if not dist.is_initialized():
        return True

    return get_rank() == 0


def synchronize():
    if not dist.is_available():
        return

    if not dist.is_initialized():
        return

    world_size = dist.get_world_size()

    if world_size == 1:
        return

    dist.barrier()


def get_world_size():
    if not dist.is_available():
        return 1

    if not dist.is_initialized():
        return 1

    return dist.get_world_size()


def reduce_sum(tensor):
    if not dist.is_available():
        return tensor

    if not dist.is_initialized():
        return tensor

    tensor = tensor.clone()
    dist.all_reduce(tensor, op=dist.ReduceOp.SUM)

    return tensor


def gather_grad(params):
    world_size = get_world_size()
    
    if world_size == 1:
        return

    for param in params:
        if param.grad is not None:
            dist.all_reduce(param.grad.data, op=dist.ReduceOp.SUM)
            param.grad.data.div_(world_size)


def all_gather(input, cat=True):
    if get_world_size() == 1:
        if cat:
            return input
        else:
            return input.unsqueeze(0)
    input_list = [torch.zeros_like(input) for _ in range(get_world_size())]
    synchronize()
    torch.distributed.all_gather(input_list, input, async_op=False)
    if cat:
        inputs = torch.cat(input_list, dim=0)
    else:
        inputs = torch.stack(input_list, dim=0)
    return inputs


def all_gatherv(input, return_boundaries=False):
    """Variable-sized all_gather"""

    # Broadcast the number of elements in every process:
    num_elements = torch.tensor(input.size(0), device=input.device)
    num_elements_per_process = all_gather(num_elements, cat=False)
    max_elements = num_elements_per_process.max()
    # Add padding so every input is the same size:
    difference = max_elements - input.size(0)
    if difference > 0:
        input = torch.cat([input, torch.zeros(difference, *input.size()[1:], device=input.device, dtype=input.dtype)], 0)
    inputs = all_gather(input, cat=False)
    # Remove padding:
    inputs = torch.cat([row[:num_ele] for row, num_ele in zip(inputs, num_elements_per_process)], 0)
    if return_boundaries:
        boundaries = torch.cumsum(num_elements_per_process, dim=0)
        boundaries = torch.cat([torch.zeros(1, device=input.device, dtype=torch.int), boundaries], 0)
        return inputs, boundaries.long()
    else:
        return inputs


def all_reduce(input, device):
    num_local = torch.tensor([input.size(0)], dtype=torch.float, device=device)
    input = input.sum(dim=0, keepdim=True).to(device)
    num_global = all_gather(num_local).sum()
    input = all_gather(input)
    input = input.sum(dim=0).div(num_global)
    return input


def rank0_to_all(input):
    input = all_gather(input)
    rank0_input = input[0]
    return rank0_input


def reduce_loss_dict(loss_dict):
    world_size = get_world_size()

    if world_size < 2:
        return loss_dict

    with torch.no_grad():
        keys = []
        losses = []

        for k in sorted(loss_dict.keys()):
            keys.append(k)
            losses.append(loss_dict[k])

        losses = torch.stack(losses, 0)
        dist.reduce(losses, dst=0)

        if dist.get_rank() == 0:
            losses /= world_size

        reduced_losses = {k: v for k, v in zip(keys, losses)}

    return reduced_losses




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_utils.ipynb.

# %% auto 0
__all__ = ['ICACHE', 'VERBOSE', 'AVCV_CACHE_DIR', 'identify', 'self_memoize', 'memoize', 'imemoize', 'is_interactive', 'get_name',
           'get_files', 'find_contours', 'mkdir', 'put_text', 'video_to_images', 'v2i', 'images_to_video', 'av_i2v',
           'md5_from_str', 'VideoReader', 'TimeLoger', 'generate_tmp_filename', 'get_md5', 'printc']

# %% ../nbs/03_utils.ipynb 1
from ._imports import *

# %% ../nbs/03_utils.ipynb 3
ICACHE = dict()

def identify(x):
    '''Return an hex digest of the input'''
    return xxhash.xxh64(pickle.dumps(x), seed=0).hexdigest()

VERBOSE = False
AVCV_CACHE_DIR = '/tmp/avcv/'

def self_memoize(func):
    """
        Decorator for a function member of an object
    """
    @wraps(func)
    def _f(self, *args, **kwargs):
        
        if not hasattr(self, 'cache'):
            self.cache = {}
        ident_name = identify(args)
        timer = mmcv.Timer()
        if ident_name in self.cache:
            result = self.cache[ident_name]
            # print('[Cached]: {:0.2f}'.format(timer.since_start()))
        else:
            result = self.cache[ident_name] = func(self, *args, **kwargs)
            # print('[No cache]: {:0.6f}'.format(timer.since_start()))
        return result
    
    return _f

def memoize(func):

    '''Cache result of function call on disk
    Support multiple positional and keyword arguments'''
    @wraps(func)
    def memoized_func(*args, **kwargs):
        try:
            if 'cache_key' in kwargs:
                cache_key = kwargs['cache_key']
                
                func_id = identify((inspect.getsource(func)))+'_cache_key_'+str(kwargs['cache_key'])
                if VERBOSE:
                    logger.info(f'Use {cache_key=}->{func_id=}')
            else:
                func_id = identify((inspect.getsource(func), args, kwargs))
            cache_path = os.path.join(AVCV_CACHE_DIR, 'funcs', func.__name__+'/'+func_id)
            mmcv.mkdir_or_exist(os.path.dirname(cache_path))

            if (os.path.exists(cache_path) and
                    not func.__name__ in os.environ and
                    not 'BUST_CACHE' in os.environ):
                result = pickle.load(open(cache_path, 'rb'))
            else:
                result = func(*args, **kwargs)
                pickle.dump(result, open(cache_path, 'wb'))
            return result
        except (KeyError, AttributeError, TypeError, Exception) as e:
            logger.warning(f'Exception: {e}, use default function call')
            return func(*args, **kwargs)
    return memoized_func

def imemoize(func):
    """
        Memoize a function into memory, the function recaculate only 
        change when its belonging arguments change
    """
    @wraps(func)
    def _f(*args, **kwargs):

        timer = mmcv.Timer()

        ident_name = identify((inspect.getsource(func), args, kwargs))
        # if not ident_name in ICACHE:
        try:
            result = ICACHE[ident_name]
        except:
            result = func(*args, **kwargs)
            ICACHE[ident_name] = result
        return result
    return _f

# %% ../nbs/03_utils.ipynb 5
def is_interactive():
    try:
        shell = get_ipython().__class__.__name__
        return True
    except:
        return False
    

def get_name(path):
    path = osp.basename(path).split('.')[:-1]
    return '.'.join(path)

def get_files(folder, ext='*'):
    return glob(osp.join(folder, ext))

def find_contours(thresh):
    """
        Get contour of a binary image
            Arguments:
                thresh: binary image
            Returns:
                Contours: a list of contour
                Hierarchy:

    """
    try:
        contours, hierarchy = cv2.findContours(thresh, cv2.RETR_TREE,
                                               cv2.CHAIN_APPROX_SIMPLE)
        return contours, hierarchy[0]
    except:
        return None, None

def mkdir(path):
    os.makedirs(path, exist_ok=True)


def put_text(image, pos, text, color=(255, 255, 255)):
    return cv2.putText(image, text, pos, cv2.FONT_HERSHEY_SIMPLEX, 1.0,
                       color, 2)


def video_to_images(input_video, output_dir=None, skip=1, rescale=None):
    """
        Extract video to image:
            inputs:
                input_video: path to video
                output_dir: default is set to video name
    """

    if output_dir is None:
        vname = get_name(input_video).split('.')[0]
        output_dir = osp.join('.cache/video_to_images/', vname)
        logger.info(f'Set output_dir = {output_dir}')

    video = mmcv.video.VideoReader(input_video)
    logger.info(f'Extracting video to images -> {output_dir}')
    pbar = mmcv.ProgressBar(video._frame_cnt)
    n_update = max(1, len(video)/1000)
    for i in range(0, len(video), skip):
        out_img_path = os.path.join(output_dir, f'{i:06d}' + '.jpg')

        if not osp.exists(out_img_path):
            try:
                img = video[i]
                if rescale is not None:
                    img = mmcv.imrescale(img, (rescale, rescale))

                mmcv.imwrite(img, out_img_path)
            except Exception as e:
                logger.warning(f"Cannot write image index {i}, exception: {e}")
                continue
        if i%n_update == 0:
            pbar.update(n_update)

@call_parse
def v2i(input_video:Param("", str), output_dir:Param("", str)=None, skip:Param("", int)=1, rescale:Param("", int)=None):
    return video_to_images(input_video, output_dir, skip, rescale)


# %% ../nbs/03_utils.ipynb 6
def images_to_video(
        images,
        out_path=None,
        fps: int = 30,
        no_sort=False,
        max_num_frame=10e12,
        resize_rate=1,
        with_text=False,
        text_is_date=False,
        verbose=True,
        resize=False,
        output_size=None
):
    assert output_size is not None
    if out_path is None:
        assert isinstance(
            images, str), "No out_path specify, you need to input a string to a directory"
        out_path = images+'.mp4'
    if isinstance(images, str) and os.path.isdir(images):
        images = glob(os.path.join(images, "*.jpg")) + \
            glob(os.path.join(images, "*.png")) + \
            glob(os.path.join(images, "*.jpeg"))

    def get_num(s):
        try:
            s = os.path.basename(s)
            num = int(''.join([c for c in s if c.isdigit()]))
        except:
            num = s
        return num
    
    def f(img_or_path):
        if isinstance(img_or_path, str):
            name = os.path.basename(img_or_path)
            img = mmcv.imread(img_or_path)
            img = cv2.resize(img, output_size)
            assert img is not None, img_or_path
            if with_text:
                if text_is_date:
                    name = name.split('.')[0].split('_')
                    f = float('{}.{}'.format(*name))
                img = put_text(img, (20, 20), name)
        else:
            img = img_or_path
        return img

    if not no_sort and isinstance(images[0], str):
        images = list(sorted(images, key=get_num))

    max_num_frame = int(max_num_frame)
    max_num_frame = min(len(images), max_num_frame)


    if output_size is None:
        h, w = mmcv.imread(images[0]).shape[:2]
        output_size = (int(w*resize_rate), int(h*resize_rate))
    if out_path.endswith('.mp4'):
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(out_path, fourcc, fps, output_size)
    elif out_path.endswith('.avi'):
        out = cv2.VideoWriter(
            out_path, cv2.VideoWriter_fourcc(*'DIVX'), fps, output_size)
    else:
        raise NotImplementedError
    images = images[:max_num_frame]
    if isinstance(images[0], str) or resize:
        if verbose:
            logger.info('Read and resize images to shape {}'.format(output_size))
        images = multi_thread(f, images, verbose=verbose)
    
    if verbose:
        logger.info(f"Write video, output_size: {output_size}")
        pbar = mmcv.ProgressBar(len(images))

    n_update = max(len(images)//1000, 1)
    for i, img in enumerate(images):
        img = cv2.resize(img, output_size)
        out.write(img)
        if i % n_update == 0 and verbose:
            pbar.update(n_update)
    if verbose:
        logger.info("-> {}".format(osp.abspath(out_path)))
        
    out.release()
@call_parse
def av_i2v(
            images: Param("Path to the images folder or list of images"),
            out_path: Param("Output output video path", str)=None,
            fps: Param("Frame per second", int) = 30,
            no_sort: Param("Sort images", bool) = False,
            max_num_frame: Param("Max num of frame", int) = 10e12,
            resize_rate: Param("Resize rate", float) = 1,
            with_text: Param("Add additional index to image when writing vidoe", bool) = False,
            text_is_date: Param("Add additional index to image when writing vidoe", bool) = False,
            verbose:Param("Print...", bool)=True,
            output_size:Param("Print...", str)=None,
        ):
    if output_size is not None:
        output_size = [int(_) for _ in output_size.split(',') if _.isdigit()]
    return images_to_video(images, out_path, fps,
                           no_sort, max_num_frame, resize_rate, with_text,
                           text_is_date,verbose,output_size=output_size,
            )


# %% ../nbs/03_utils.ipynb 9
def md5_from_str(s:str):
    import hashlib
    return hashlib.md5(s.encode('utf-8')).hexdigest()

class VideoReader:
    def __init__(self, file_name, scale=0.5, mode='gray', verbose=True):
        assert mode in ['rgb', 'gray']
        self.video = mmcv.VideoReader(file_name)
        hash_name = md5_from_str(osp.abspath(file_name))
        cache_file = osp.join(AVCV_CACHE_DIR, 'VideoReader/{}_{}.array'.format(hash_name, get_name(file_name)))
        
        self.height = int(self.video.height*scale)
        self.width = int(self.video.width*scale)
        self.mode = mode
        if osp.exists(cache_file):
            self._imgs = np.memmap(cache_file, mode='r', shape=(len(self.video), self.height, self.width, 1 if self.mode == 'gray' else 3))
        else:
            os.makedirs(osp.dirname(cache_file), exist_ok=True)
            try:
                self._imgs = np.memmap(cache_file, mode='w+', shape=(len(self.video), self.height, self.width, 1 if self.mode == 'gray' else 3))
                print('Expected size: {:0.2f} gb'.format(self._imgs.size/1024/1024/1024))
                pbar = range(len(self))
                if verbose: pbar = tqdm(pbar)
                for i in pbar:
                    img = self.video[i]
                    img = mmcv.imresize(img, (self.width, self.height))
                    if mode == 'gray' and len(img.shape) == 3:
                        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)[...,None]
                    else:
                        img = img[...,::-1].copy()
                    self._imgs[i] = img
            except Exception as e:
                print(e)
                os.remove(cache_file)
        self.cache_file = cache_file
        
    def __len__(self):
        return len(self.video)
    
    def __getitem__(self, idx):
        return self._imgs[idx]
    

# %% ../nbs/03_utils.ipynb 15
class TimeLoger:
    def __init__(self):
        self.timer = mmcv.Timer()
        self.time_dict = dict()

    def start(self):
        self.timer.start()

    def update(self, name):
        # assert not name in self.time_dict
        duration = self.timer.since_last_check()
        if name in self.time_dict:
            self.time_dict[name].append(duration)
        else:
            self.time_dict[name] = [duration]

    def __str__(self):
        total_time = np.sum([np.sum(v) for v in self.time_dict.values()])
        s = f"------------------Time Loger Summary : Total {total_time:0.2f} ---------------------:\n"
        for k, v in self.time_dict.items():
            average = np.mean(v)
            times = len(v)
            percent = np.sum(v)*100/total_time
            s += f'\t\t{k}:  \t\t{percent:0.2f}% ({average:0.4f}s) | Times: {times} \n'
        return s

# %% ../nbs/03_utils.ipynb 16
def generate_tmp_filename():
    return tempfile.NamedTemporaryFile().name 
@memoize
def get_md5(video_path, os_system='linux'):
    tmp = tempfile.NamedTemporaryFile().name
    if os_system == 'linux':
        cmd = f'md5sum {video_path} > {tmp}'
        os.system(cmd)
        md5 = open(tmp).readlines()[0].split(' ')[0]
    else:
        os.system(f'md5 {video_path} > {tmp}')
        md5 = open(tmp).readlines()[0].split(' = ')[1][:-1]
    return md5


# %% ../nbs/03_utils.ipynb 20
# DB =  ipdb.set_trace

# %% ../nbs/03_utils.ipynb 21
def printc(module_or_func, verbose=True, return_lines=False):
    """
        Print code given a 
    """
    lines = inspect.getsource(module_or_func)
    if verbose:
        print(lines)
    if return_lines:
        return lines



# import __main__ as main
# def is_interactive():
#     return not hasattr(main, '__file__')
import inspect
import json
import os
import os.path as osp
import pickle
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from functools import partial, wraps
from glob import glob
from multiprocessing import Pool

import xxhash
from fastcore.all import *
from fastcore.parallel import threaded
from fastcore.script import *
from fastcore.script import Param, call_parse
from loguru import logger
from PIL import Image
# from pycocotools.coco import COCO
from tqdm import tqdm
# from .lazy_modules import LazyModule
from lazy_module.core import LazyModule
import copy
mmcv = LazyModule('mmcv')
np = LazyModule('numpy')
cv2 = LazyModule('cv2')
matplotlib = LazyModule('matplotlib')
plt = LazyModule('plt', 'import matplotlib.pyplot as plt')
coco = LazyModule('coco', 'from pycocotools import coco')
ipdb = LazyModule('ipdb')
pd = LazyModule('pandas')
#import ipdb, cv2, matplotlib.pyplot as plt, mmcv, numpy as np, matplotlib



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/05_coco_dataset.ipynb.

# %% auto 0
__all__ = ['PYTHON_VERSION', 'AvCOCO', 'CocoDataset', 'get_bboxes', 'get_overlap_rate', 'bbox_expand', 'DiagnoseCoco', 'to_jpg',
           'video_to_coco', 'split_coco', 'concat_coco', 'check_save_coco_dict', 'extract_coco', 'v2c']

# %% ../nbs/05_coco_dataset.ipynb 2
from ._imports import *

# %% ../nbs/05_coco_dataset.ipynb 3
PYTHON_VERSION = 3


class AvCOCO(coco.COCO):
    def __init__(self, annotation_file=None, verbose=False):
        """
        Constructor of Microsoft COCO helper class for reading and visualizing annotations.
        :param annotation_file (str): location of annotation file
        :param image_folder (str): location to the folder that hosts images.
        :return:
        """
        # load dataset
        self.dataset,self.anns,self.cats,self.imgs = dict(),dict(),dict(),dict()
        self.imgToAnns, self.catToImgs = defaultdict(list), defaultdict(list)
        self.verbose = verbose
        if not annotation_file == None:
            if verbose:
                logger.info('loading annotations into memory...')
            tic = time.time()
            if isinstance(annotation_file, str):
                with open(annotation_file, 'r') as f:
                    dataset = json.load(f)
            else:
                dataset = annotation_file
            assert type(dataset)==dict, 'annotation file format {} not supported'.format(type(dataset))
            if verbose:
                logger.info('Done (t={:0.2f}s)'.format(time.time()- tic))
            self.dataset = dataset
            self.createIndex()
            
            
    def createIndex(self):
        # create index
        if self.verbose:
            logger.info('creating index...')
        anns, cats, imgs = {}, {}, {}
        imgToAnns,catToImgs = defaultdict(list),defaultdict(list)
        if 'annotations' in self.dataset:
            for ann in self.dataset['annotations']:
                imgToAnns[ann['image_id']].append(ann)
                anns[ann['id']] = ann

        if 'images' in self.dataset:
            for img in self.dataset['images']:
                imgs[img['id']] = img

        if 'categories' in self.dataset:
            for cat in self.dataset['categories']:
                cats[cat['id']] = cat

        if 'annotations' in self.dataset and 'categories' in self.dataset:
            for ann in self.dataset['annotations']:
                catToImgs[ann['category_id']].append(ann['image_id'])
        if self.verbose:
            logger.info('index created!')

        # create class members
        self.anns = anns
        self.imgToAnns = imgToAnns
        self.catToImgs = catToImgs
        self.imgs = imgs
        self.cats = cats

    def loadRes(self, resFile):
        """
        Load result file and return a result api object.
        :param   resFile (str)     : file name of result file
        :return: res (obj)         : result api object
        """
        res = coco.COCO()
        res.dataset['images'] = [img for img in self.dataset['images']]
        if self.verbose:
            logger.info('Loading and preparing results...')
        tic = time.time()
        if type(resFile) == str or (PYTHON_VERSION == 2 and type(resFile) == unicode):
            with open(resFile) as f:
                anns = json.load(f)
        elif type(resFile) == np.ndarray:
            anns = self.loadNumpyAnnotations(resFile)
        else:
            anns = resFile
        assert type(anns) == list, 'results in not an array of objects'
        annsImgIds = [ann['image_id'] for ann in anns]
        assert set(annsImgIds) == (set(annsImgIds) & set(self.getImgIds())), \
               'Results do not correspond to current coco set'
        if len(anns):
            if 'caption' in anns[0]:
                imgIds = set([img['id'] for img in res.dataset['images']]) & set([ann['image_id'] for ann in anns])
                res.dataset['images'] = [img for img in res.dataset['images'] if img['id'] in imgIds]
                for id, ann in enumerate(anns):
                    ann['id'] = id+1
            elif 'bbox' in anns[0] and not anns[0]['bbox'] == []:
                res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
                for id, ann in enumerate(anns):
                    bb = ann['bbox']
                    x1, x2, y1, y2 = [bb[0], bb[0]+bb[2], bb[1], bb[1]+bb[3]]
                    if not 'segmentation' in ann:
                        ann['segmentation'] = [[x1, y1, x1, y2, x2, y2, x2, y1]]
                    ann['area'] = bb[2]*bb[3]
                    ann['id'] = id+1
                    ann['iscrowd'] = 0
            elif 'segmentation' in anns[0]:
                res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
                for id, ann in enumerate(anns):
                    # now only support compressed RLE format as segmentation results
                    ann['area'] = maskUtils.area(ann['segmentation'])
                    if not 'bbox' in ann:
                        ann['bbox'] = maskUtils.toBbox(ann['segmentation'])
                    ann['id'] = id+1
                    ann['iscrowd'] = 0
            elif 'keypoints' in anns[0]:
                res.dataset['categories'] = copy.deepcopy(self.dataset['categories'])
                for id, ann in enumerate(anns):
                    s = ann['keypoints']
                    x = s[0::3]
                    y = s[1::3]
                    x0,x1,y0,y1 = np.min(x), np.max(x), np.min(y), np.max(y)
                    ann['area'] = (x1-x0)*(y1-y0)
                    ann['id'] = id + 1
                    ann['bbox'] = [x0,y0,x1-x0,y1-y0]
            if self.verbose:
                logger.info('DONE (t={:0.2f}s)'.format(time.time()- tic))

        res.dataset['annotations'] = anns
        res.createIndex()
        return res        
        
        


# %% ../nbs/05_coco_dataset.ipynb 5
from .visualize import bbox_visualize, show as av_show
#import COCOeval


class CocoDataset:
    def __init__(self, gt, img_dir=None, pred=None, verbose=False):
        if img_dir is None:
            assert isinstance(gt, str) and '/annotations/' in gt
            img_dir = gt.split('/annotations/')[0]+'/images'
            if verbose:
                logger.warning(f'Img dir is not set, set to :{img_dir}')
            # assert osp.isdir(img_dir)
            
        if isinstance(gt, coco.COCO):
            gt = gt.dataset
        self.gt = AvCOCO(gt, verbose=verbose)

        if isinstance(pred, str):
            pred = mmcv.load(pred)

        self.pred = self.gt.loadRes(pred) if pred is not None else None

        self.img_dir = img_dir
        self.img_ids = [_['id'] for _ in self.gt.imgs.values()]

    def imread(self, img_id, channel_order='bgr'):
        im = self.gt.imgs[img_id]
        img_path = osp.abspath(osp.join(self.img_dir,im['file_name']))
        assert osp.exists(img_path), img_path
        return mmcv.imread(img_path, channel_order=channel_order)

    def visualize(self, img_id=None,  mode='gt', dpi=100, 
        show=False, anns=None, color='green', img=None, score_thr=0.3, box_color=None):
        if img_id is None:
            img_id = np.random.choice(self.img_ids)
            logger.info(f'Random visualize img_id={img_id}')
        if img is None:
            img= self.imread(img_id)

        if mode=='pred':
            assert self.pred is not None

        CLASSES = {cat_id:cat['name'] for cat_id, cat in self.gt.cats.items()}
        im = self.gt.imgs[img_id]
        source = self.pred if mode == 'pred' else self.gt

        if anns is None:
            anns  = source.loadAnns(source.getAnnIds(im['id']))

        bboxes = []
        lables = []
        scores = []
        for ann in anns:
            x1,y1,w,h = [int(_) for _ in ann['bbox']]
            x2 = x1+w
            y2 = y1+h
            scores.append(ann.get('score', 1))
            bboxes.append([x1,y1,x2,y2])
            lables.append(ann['category_id'])

        bboxes = np.array(bboxes)
        lables = np.array(lables)

        if len(bboxes):
            
            img = bbox_visualize(img, bboxes, scores, lables, score_thr, CLASSES, box_color=box_color)
#             img = mmcv.visualization.imshow_det_bboxes(img, bboxes, 
#                 lables, CLASSES, show=False, bbox_color=color, text_color=color, 
#                 score_thr=score_thr)
        if show:
            av_show(img[...,::-1], dpi=dpi)
        return img

    def load_anns(self, img_id, source=None):
        if source is None:
            source = self.gt
        anns = source.loadAnns(source.getAnnIds(img_id))
        return anns

    def evaluate(self, *args, **kwargs):
        from pycocotools.cocoeval import COCOeval
        cocoEval = COCOeval(self.gt, self.pred, 'bbox')
        cocoEval.evaluate()
        cocoEval.accumulate()
        cocoEval.summarize()


# %% ../nbs/05_coco_dataset.ipynb 9
def get_bboxes(anns,category_ids,
               mode='xywh',
               dtype=np.float32,
               score_thr=None, with_score=False):
    bboxes = []
    for ann in anns:
        if category_ids is not None and  not ann['category_id'] in category_ids: continue
        if score_thr is not None and ann.get('score', False):
            if ann['score'] < score_thr:
                continue
        x, y, w, h = ann['bbox']
        
        score = ann.get('score', 1)
        if mode == 'xywh':
            bboxes.append([x, y, w, h, score])
        elif mode == 'xyxy':
            bboxes.append([x, y, x + w, y + h, score])
        elif mode == 'cxcywh':
            cx = x + w / 2
            cy = y + h / 2
            bboxes.append([cx, cy, w, h, score])
        else:
            raise NotImplemented
    bboxes = np.array(bboxes).reshape([-1, 5])
    if not with_score:
        bboxes = bboxes[:,:4]
        
    if dtype is not None:
        bboxes = bboxes.astype(dtype)
    return bboxes


def get_overlap_rate(boxA, boxB):
    xA = max(boxA[0], boxB[0])
    yA = max(boxA[1], boxB[1])
    xB = min(boxA[2], boxB[2])
    yB = min(boxA[3], boxB[3])
    interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)
    boxAArea = (boxA[2] - boxA[0] + 1) * (boxA[3] - boxA[1] + 1)
    iou = interArea / float(boxAArea)
    return iou


def bbox_expand(bb, im_h, im_w, r=1.25):
    x,y,w,h = bb
    cx = x+w/2
    cy = y+h/2
    h*=r
    w*=r
    
    x1 = cx-w/2
    y1 = cy-h/2
    x2 = x1+w
    y2 = y1+h
    
    x1 = max(0, x1)
    y1 = max(0, y1)
    x2 = min(im_w, x2)
    y2 = min(im_h, y2)
    
    return x1, y1, x2-x1, y2-y1


# %% ../nbs/05_coco_dataset.ipynb 11
class DiagnoseCoco(CocoDataset):
    COLORS = dict(
        TP=(0, 255, 0),
        FN=(0, 0, 255), # Undetected GT
        FP=(0,255,255), # Wrong detection
    )

    def find_false_samples(self, img_id, score_thr=0.05, visualize=True, device='cuda:7'):
        import torch
        # from cython_bbox import bbox_overlaps
        from mmcv.ops import bbox_overlaps


        assert self.gt is not None
        assert self.pred is not None
        result = dict(tp=[], fn=[], fp=[])
        pred_anns = [ann for ann in self.pred.loadAnns(self.pred.getAnnIds(img_id)) if ann['score']>score_thr]
        gt_anns = self.gt.loadAnns(self.gt.getAnnIds(img_id))
        if len(pred_anns) == 0 and len(gt_anns)==0:
            return result
        
        pred_bboxes = get_bboxes(pred_anns,None , mode='xyxy')
        gt_bboxes = get_bboxes(gt_anns, None,mode='xyxy')
        
        with torch.no_grad():
            pred_bboxes = torch.from_numpy(pred_bboxes).to(device)
            gt_bboxes = torch.from_numpy(gt_bboxes).to(device)
            ious = bbox_overlaps(pred_bboxes, gt_bboxes).cpu().numpy()
            
        mapping_gt_pred = np.where(ious>0)

        

        gt_ids = list(range(len(gt_anns)))
        pred_ids = list(range(len(pred_anns)))
        for pred_id, gt_id in zip(*mapping_gt_pred):
            if gt_anns[gt_id]['category_id'] == pred_anns[pred_id]['category_id']:
                result['tp'].append(pred_anns[pred_id])
                if gt_id in gt_ids:
                    gt_ids.remove(gt_id)
                if pred_id in pred_ids:
                    pred_ids.remove(pred_id)


        result['fp'] = [pred_anns[i] for i in  pred_ids]
        result['fn'] = [gt_anns[i] for i in gt_ids]
        if visualize:
            vis_img = self.visualize(img_id, anns=result['fn'], color=self.COLORS['FN'], show=False, box_color=self.COLORS['FN'])
            vis_img = self.visualize(img_id,  anns=result['tp'],img=vis_img, show=False,box_color=self.COLORS['TP'])
            vis_img = self.visualize(img_id, anns=result['fp'], dpi=150,color=self.COLORS['FP'], show=False,
                                    img=vis_img, box_color=self.COLORS['FP'])
            vis_img = vis_img[...,::-1].copy()
            result['vis_img'] = vis_img
        return result


# %% ../nbs/05_coco_dataset.ipynb 13
def _f(p):
    img = mmcv.imread(p)
    mmcv.imwrite(img, p.replace('.png', '.jpg'))
    os.remove(p)
    
def to_jpg(img_dir):
    paths = glob(osp.join(img_dir, '*.png'))
    logger.info(f'Converting folder {img_dir} to jpg...')
    with Pool(10) as p:
        it = p.imap(_f, paths)
        total = len(paths)
        r = list(tqdm(it, total=total))

def video_to_coco(
    input_video,
    test_json,
    output_dir=None,
    skip=1,
    rescale=None,
    recursive=False,
    min_is_done_extracted_ratio=.7,
):

    assert os.path.exists(input_video), f'{input_video} does not exist'
    try:
        fps = mmcv.VideoReader(input_video).fps
    except:
        fps = None
        
    def path2image(path, root_dir):
        w, h = Image.open(path).size
        name = path.replace(root_dir, '')
        if name.startswith('/'):
            name = name[1:]
        return dict(
            file_name=name, height=h, width=w
        )
    

    if output_dir is None:
        name  = get_name(input_video) if not osp.isdir(input_video) else \
            os.path.normpath(input_video).split('/')[-1]
        output_dir = osp.join('.cache/video_to_coco', name)
        logger.info(f'Set output dir to->{output_dir}')

    image_out_dir = osp.join(output_dir, 'images')

    image_dir_name = osp.normpath(image_out_dir).split('/')[-1]
    path_out_json = osp.join(output_dir, f'annotations/{image_dir_name}.json')

    mmcv.mkdir_or_exist(osp.dirname(path_out_json))

    source_type = 'dir' if osp.isdir(input_video) else 'video'
    
    if not osp.isdir(input_video): # IF VIDEO => EXTRACT IMAGES to image_out_dir
        mmcv.mkdir_or_exist(image_out_dir)
        video = mmcv.VideoReader(input_video)
        

        if rescale is not None:
            _im = mmcv.imrescale(video[0], rescale)
            im_h, im_w = _im.shape[:2]#int(rescale*im_h), int(rescale*im_w)
        else:
            im_h, im_w = video[0].shape[:2]
        num_found_imgs = len(glob(osp.join(image_out_dir, '*.jpg')))
        is_done_extracted = (num_found_imgs / len(video)) > min_is_done_extracted_ratio
        if not num_found_imgs  == len(video):
            logger.warning(f'Differ num of extracted images and video len {num_found_imgs=}, {len(video)=}, {is_done_extracted=}')
        
        
        if not is_done_extracted:
            mmcv.mkdir_or_exist(image_out_dir)
            cmd = f"ffmpeg  -i {input_video} -s {im_w}x{im_h} {image_out_dir}/%06d.png"
            logger.info(f'Running command: {cmd}')
            os.system(cmd)
            to_jpg(image_out_dir)
        else:
            logger.info('Skip extracting video')
        
    else:
        # IF FOLDER THEN CREATE SYMLINK
        logger.info(f'Symn link {input_video}-> {image_out_dir}')
        os.symlink(osp.abspath(input_video), osp.abspath(image_out_dir))
        
    paths = list(sorted(glob(osp.join(image_out_dir, '*.jpg'), recursive=recursive)))
    paths += list(sorted(glob(osp.join(image_out_dir, '*.png'), recursive=recursive)))
    out_dict = dict(images=[], annotations=[], meta=dict(fps=fps),
                    categories=mmcv.load(test_json)['categories'])
    out_dict['images'] = list(
        map(partial(path2image, root_dir=image_out_dir), paths))

    for i, image in enumerate(out_dict['images']):
        image['id'] = i
    mmcv.dump(out_dict, path_out_json)
    return os.path.normpath(path_out_json), os.path.normpath(image_out_dir)




# %% ../nbs/05_coco_dataset.ipynb 16
def split_coco(coco, train_ratio=0.85, seed=0):
    if isinstance(coco, dict):
        coco = AvCOCO(coco)
    ids = list(coco.imgs.keys())
    np.random.seed(seed)
    np.random.shuffle(ids)
    train_n = int(len(ids)*train_ratio)
    train_ids = ids[:train_n]
    test_ids = ids[train_n:]
    logger.info('Spliting ratio {} -> {} training and {} testing'.format(train_ratio, len(train_ids), len(test_ids)))
    
    # train_ds = coco.dataset.copy()
    # val_ds = coco.dataset.copy()
    def get_ds(ids):
        out = dict(categories=coco.dataset['categories'], images=[], annotations=[])
        for img_id in ids:
            out['images'].append(coco.imgs[img_id])
            out['annotations'].extend(coco.imgToAnns[img_id])
        return out
    return get_ds(train_ids), get_ds(test_ids)


def concat_coco(datasets, new_root, name=None, cat_name2id=None, categories=None, strict_image_path=True, identical_cat_name=True):
    """
        Example:
            concat_coco([
                (train_ds_1, '/data/full-version-vip-pro/DMS_DB_090922/'),
                (train_ds_2, '/data/DMS_Behavior_Detection/merge-phone-cigaret-food/images'),

            ], '/data/face_food_concat/', 'train');
        """
    # If datasets if a list of string
    if isinstance(datasets[0], str):
        inputs = []
        # for j in datasets:
        def f(j):
            cc = CocoDataset(j)
            return (cc.gt.dataset, cc.img_dir)
        from avcv.all import multi_thread
        datasets = multi_thread(f, datasets)
            
    if cat_name2id is None:
        if identical_cat_name:
            av_coco = AvCOCO(datasets[0][0])
            cat_name2id = {cat['name']:cat['id'] for cat in av_coco.cats.values()}
        else:
            all_cat_name = []
            for dataset in datasets:
                av_coco = AvCOCO(dataset[0])
                all_cat_name.extend(
                    [cat['name'].lower() for cat in av_coco.cats.values()])

            all_cat_name = list(set(all_cat_name))
            cat_name2id = {name: i + 1 for i, name in enumerate(all_cat_name)}
        print(f'cat_name2id={cat_name2id}')

    mmcv.mkdir_or_exist(osp.join(new_root, 'annotations'))
    out_concat = dict(images=[],
                      annotations=[],
                      categories=[
                          dict(name=name, id=id)
                          for name, id in cat_name2id.items()
                      ] if categories is None else categories)
    CACHE_IMAGE_PATH2ID = {}
    def get_image_id(image_path):
        image_path = osp.abspath(image_path)
        if not image_path in CACHE_IMAGE_PATH2ID:
            CACHE_IMAGE_PATH2ID[image_path] = len(CACHE_IMAGE_PATH2ID)+1
            
        return CACHE_IMAGE_PATH2ID[image_path]
        
    added_image_ids = set()
    
    for i, inp in enumerate(datasets):
        if not isinstance(inp, tuple):
            datasets[i] = (inp, inp.split('/annotations/')[0]+'/images')
    
    for json_path, old_img_dir in datasets:

        if isinstance(json_path, str) or isinstance(json_path, dict):
            dataset = AvCOCO(json_path)
        else :
            dataset = json_path
        for img_id in dataset.imgs:
            image = dataset.imgs[img_id]
            anns = dataset.imgToAnns[
                image['id']] if img_id in dataset.imgToAnns else []

            old_path = osp.join(old_img_dir, image['file_name'])
            old_path = osp.abspath(old_path)
            if strict_image_path:
                assert osp.exists(old_path), old_path
            image['file_name'] = osp.relpath(old_path,
                                             osp.join(new_root, 'images'))
            
            image['id'] = get_image_id(old_path)
            for ann in anns:
                cat_name = dataset.cats[ann['category_id']]['name']
                ann['category_id'] = cat_name2id[cat_name.lower()]
                ann['image_id'] = image['id']
                w, h = ann['bbox'][-2:]
                ann['area'] = w * h
                ann['id'] = len(out_concat['annotations'])

                out_concat['annotations'].append(ann)
                
            if not image['id'] in added_image_ids:
                valid_classes = set([cat['name'] for cat in dataset.cats.values()]) if not 'pred_' in json_path else None
                image['valid_classes'] = valid_classes
                out_concat['images'].append(image)
                added_image_ids.add(image['id'])

    if name is not None:
        out_json_path = osp.join(new_root, f'annotations/{name}.json')
        logger.info('Annotation is saved to: {}', out_json_path)
        check_save_coco_dict(out_concat)
        mmcv.dump(out_concat, out_json_path)
    # Summary num of images, annotations and categories
    logger.info('Summary:'
    ' {} images, {} annotations'.format(len(out_concat['images']), len(out_concat['annotations']))
    )
    return out_concat


def check_save_coco_dict(data):
    def check_duplicated_id(name):
        ids = set()
        for item in data[name]:
            if not item['id'] in ids:
                ids.add(item['id'])
            else:
                print(f'{name}, Id: {id} is duplicated, item: {item}')
                return False
        return True
    if check_duplicated_id('images') and check_duplicated_id('annotations') and check_duplicated_id('categories'):
        print('This dict is cocoly safe')
        return True
    return False

def extract_coco(coco, img_ids):
    if isinstance(coco, dict) or isinstance(coco, str):
        coco = AvCOCO(coco)
    imgs = coco.loadImgs(img_ids)
    anns = coco.loadAnns(coco.getAnnIds(img_ids))
    return dict(images=imgs, annotations=anns, categories=coco.dataset['categories'])


# %% ../nbs/05_coco_dataset.ipynb 19
@call_parse
def v2c(input_video: Param("path to video", str),
        test_json: Param("path to annotation json path, to get the category", str),
        output_dir: Param("", str) = None,
        skip: Param("", int) = 1,        rescale: Param("", int) = None, recursive: Param("Images recursive", bool)=False
        ):
    return video_to_coco(input_video, test_json, output_dir, skip, rescale=rescale, recursive=recursive)




# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"plot_images": "00_visualize.ipynb",
         "imshow": "00_visualize.ipynb",
         "show": "00_visualize.ipynb",
         "tensor2imgs": "00_visualize.ipynb",
         "bbox_visualize": "00_visualize.ipynb",
         "multi_thread": "01_process.ipynb",
         "identify": "03_memoize.ipynb",
         "memoize": "03_memoize.ipynb",
         "imemoize": "03_memoize.ipynb",
         "ICACHE": "03_memoize.ipynb",
         "VERBOSE": "03_memoize.ipynb",
         "CACHE_DIR": "03_memoize.ipynb",
         "is_interactive": "03_utils.ipynb",
         "get_name": "03_utils.ipynb",
         "get_files": "03_utils.ipynb",
         "find_contours": "03_utils.ipynb",
         "download_file_from_google_drive": "03_utils.ipynb",
         "mkdir": "03_utils.ipynb",
         "put_text": "03_utils.ipynb",
         "video_to_images": "03_utils.ipynb",
         "v2i": "03_utils.ipynb",
         "images_to_video": "03_utils.ipynb",
         "av_i2v": "03_utils.ipynb",
         "TimeLoger": "03_utils.ipynb",
         "generate_tmp_filename": "03_utils.ipynb",
         "get_md5": "03_utils.ipynb",
         "printc": "03_utils.ipynb",
         "make_mini_coco": "04_debug.ipynb",
         "dpython": "04_debug.ipynb",
         "AvCOCO": "05_coco_dataset.ipynb",
         "PYTHON_VERSION": "05_coco_dataset.ipynb",
         "CocoDataset": "05_coco_dataset.ipynb",
         "get_bboxes": "05_coco_dataset.ipynb",
         "get_overlap_rate": "05_coco_dataset.ipynb",
         "bbox_expand": "05_coco_dataset.ipynb",
         "DiagnoseCoco": "05_coco_dataset.ipynb",
         "to_jpg": "05_coco_dataset.ipynb",
         "video_to_coco": "05_coco_dataset.ipynb",
         "split_coco": "05_coco_dataset.ipynb",
         "concat_coco": "05_coco_dataset.ipynb",
         "check_save_coco_dict": "05_coco_dataset.ipynb",
         "extract_coco": "05_coco_dataset.ipynb",
         "v2c": "05_coco_dataset.ipynb",
         "convert_image": "06_cli.ipynb",
         "setup_distributed": "07_dist_utils.ipynb",
         "get_rank": "07_dist_utils.ipynb",
         "primary": "07_dist_utils.ipynb",
         "synchronize": "07_dist_utils.ipynb",
         "get_world_size": "07_dist_utils.ipynb",
         "reduce_sum": "07_dist_utils.ipynb",
         "gather_grad": "07_dist_utils.ipynb",
         "all_gather": "07_dist_utils.ipynb",
         "all_gatherv": "07_dist_utils.ipynb",
         "all_reduce": "07_dist_utils.ipynb",
         "rank0_to_all": "07_dist_utils.ipynb",
         "reduce_loss_dict": "07_dist_utils.ipynb",
         "__all__": "all.ipynb"}

modules = ["visualize.py",
           "process.py",
           "utils.py",
           "debug.py",
           "coco.py",
           "cli.py",
           "dist_utils.py",
           "all.py"]

doc_url = "https://anhvth.github.io/avcv/"

git_url = "https://github.com/anhvth/avcv/tree/nbdev-convert/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_visualize.ipynb.

# %% auto 0
__all__ = ['show', 'COCO_COLORS', 'plot_images', 'imshow', 'tensor2imgs', 'bbox_visualize', 'Board']

# %% ../nbs/00_visualize.ipynb 3
from ._imports import *

# %% ../nbs/00_visualize.ipynb 4
def plot_images(images,
                labels=None,
                cls_true=None,
                cls_pred=None,
                space=(0.3, 0.3),
                mxn=None,
                size=(5, 5),
                dpi=300,
                max_w=1500,
                out_file=None,
                cmap='binary'):

    if mxn is None:
        # n = max(max_w // max([img.shape[1] for img in images]), 1)
        n = int(np.sqrt(len(images)))
        n = min(n, len(images))
        m = len(images) // n
        m = max(1, m)
        mxn = (m, n)
        logger.info(f"Grid size: {mxn}")

    fig, axes = plt.subplots(*mxn)
    fig.subplots_adjust(hspace=space[0], wspace=space[1])
    fig.figsize = size
    fig.dpi = dpi
    for i, ax in enumerate(axes.flat):
        if i < len(images):
            ax.imshow(images[i], cmap=cmap)
            if labels is not None:
                xlabel = labels[i]
            elif cls_pred is None and cls_true is not None:
                xlabel = "True: {0}".format(cls_true[i])
            elif cls_pred is None and cls_true is not None:
                xlabel = "True: {0}, Pred: {1}".format(cls_true[i],
                                                       cls_pred[i])
            else:
                xlabel = None
            if xlabel is not None:
                ax.set_xlabel(xlabel)
            ax.set_xticks([])
            ax.set_yticks([])
    if out_file is None:
        plt.show()
    else:
        plt.savefig(out_file)
        logger.info('Save fig:', out_file)
        plt.close()
def imshow(inp,  dpi=100, size=10, cmap='gray', out_file=None):
    """
        Input: either a path or image
    """
    # inp = mmcv.imread(inp)
    if len(inp.shape) == 4:
        inp = inp[0]
    inp = np.squeeze(inp)
    if type(inp) is str:
        assert os.path.exists(inp)
        inp = cv2.imread(inp)
    if size is None:
        size = max(5, inp.shape[1] // 65)
    plt.figure(figsize=(size, size), dpi=dpi)
    plt.imshow(inp, cmap=cmap)
    if out_file is None:
        plt.show()
    else:
        plt.savefig(out_file)
    plt.close()
show = imshow

# %% ../nbs/00_visualize.ipynb 9
def tensor2imgs(tensor, mode='bhwc', 
                    mean=(123.675, 116.28, 103.53), std= (58.395, 57.120000000000005, 57.375), **kwargs):
    tensor = tensor.cpu()
    if mode == 'bhwc':
        tensor = tensor.permute([0,3,1,2])
        return tensor2imgs(tensor, mode='bchw',std=std, mean=mean)
    if mode == 'hwc':
        tensor = tensor[None].permute([0,3,1,2])
        return tensor2imgs(tensor, mode='bchw', std=std, mean=mean)[0]
    if mode == 'chw':
        tensor = tensor[None]
        return tensor2imgs(tensor, mode='bchw', std=std, mean=mean)[0]
    return mmcv.tensor2imgs(tensor, mean=mean, std=std, **kwargs)

# %% ../nbs/00_visualize.ipynb 12
COCO_COLORS = np.array(
    [
        0.000, 0.447, 0.741,
        0.850, 0.325, 0.098,
        0.929, 0.694, 0.125,
        0.494, 0.184, 0.556,
        0.466, 0.674, 0.188,
        0.301, 0.745, 0.933,
        0.635, 0.078, 0.184,
        0.300, 0.300, 0.300,
        0.600, 0.600, 0.600,
        1.000, 0.000, 0.000,
        1.000, 0.500, 0.000,
        0.749, 0.749, 0.000,
        0.000, 1.000, 0.000,
        0.000, 0.000, 1.000,
        0.667, 0.000, 1.000,
        0.333, 0.333, 0.000,
        0.333, 0.667, 0.000,
        0.333, 1.000, 0.000,
        0.667, 0.333, 0.000,
        0.667, 0.667, 0.000,
        0.667, 1.000, 0.000,
        1.000, 0.333, 0.000,
        1.000, 0.667, 0.000,
        1.000, 1.000, 0.000,
        0.000, 0.333, 0.500,
        0.000, 0.667, 0.500,
        0.000, 1.000, 0.500,
        0.333, 0.000, 0.500,
        0.333, 0.333, 0.500,
        0.333, 0.667, 0.500,
        0.333, 1.000, 0.500,
        0.667, 0.000, 0.500,
        0.667, 0.333, 0.500,
        0.667, 0.667, 0.500,
        0.667, 1.000, 0.500,
        1.000, 0.000, 0.500,
        1.000, 0.333, 0.500,
        1.000, 0.667, 0.500,
        1.000, 1.000, 0.500,
        0.000, 0.333, 1.000,
        0.000, 0.667, 1.000,
        0.000, 1.000, 1.000,
        0.333, 0.000, 1.000,
        0.333, 0.333, 1.000,
        0.333, 0.667, 1.000,
        0.333, 1.000, 1.000,
        0.667, 0.000, 1.000,
        0.667, 0.333, 1.000,
        0.667, 0.667, 1.000,
        0.667, 1.000, 1.000,
        1.000, 0.000, 1.000,
        1.000, 0.333, 1.000,
        1.000, 0.667, 1.000,
        0.333, 0.000, 0.000,
        0.500, 0.000, 0.000,
        0.667, 0.000, 0.000,
        0.833, 0.000, 0.000,
        1.000, 0.000, 0.000,
        0.000, 0.167, 0.000,
        0.000, 0.333, 0.000,
        0.000, 0.500, 0.000,
        0.000, 0.667, 0.000,
        0.000, 0.833, 0.000,
        0.000, 1.000, 0.000,
        0.000, 0.000, 0.167,
        0.000, 0.000, 0.333,
        0.000, 0.000, 0.500,
        0.000, 0.000, 0.667,
        0.000, 0.000, 0.833,
        0.000, 0.000, 1.000,
        0.000, 0.000, 0.000,
        0.143, 0.143, 0.143,
        0.286, 0.286, 0.286,
        0.429, 0.429, 0.429,
        0.571, 0.571, 0.571,
        0.714, 0.714, 0.714,
        0.857, 0.857, 0.857,
        0.000, 0.447, 0.741,
        0.314, 0.717, 0.741,
        0.50, 0.5, 0
    ]
).astype(np.float32).reshape(-1, 3)

def bbox_visualize(img, boxes, scores, cls_ids, conf=0.5, class_names=None, texts=None, box_color=None):
    img = img.copy()
    for i in range(len(boxes)):
        box = boxes[i]
        cls_id = int(cls_ids[i])
        score = scores[i]
        if score < conf:
            continue
        x0 = int(box[0])
        y0 = int(box[1])
        x1 = int(box[2])
        y1 = int(box[3])
        if box_color is None:
            
            color_id = cls_id % len(COCO_COLORS)
            _color = COCO_COLORS[color_id]
        else:
            _color = np.array(box_color)/255.
            
        color = (_color * 255).astype(np.uint8).tolist()

        txt_color = (0, 0, 0) if np.mean(_color) > 0.5 else (255, 255, 255)
        txt_bk_color = (_color * 255 * 0.7).astype(np.uint8).tolist()

            
            
        if texts is None:
            text = '{}:{:.1f}%'.format(class_names[cls_id], score * 100)
        else:
            text = texts[i]
        
        font = cv2.FONT_HERSHEY_SIMPLEX

        txt_size = cv2.getTextSize(text, font, 0.4, 1)[0]
        cv2.rectangle(img, (x0, y0), (x1, y1), color, 2)

        
        cv2.rectangle(
            img,
            (x0, y0 + 1),
            (x0 + txt_size[0] + 1, y0 + int(1.5*txt_size[1])),
            txt_bk_color,
            -1
        )
        cv2.putText(img, text, (x0, y0 + txt_size[1]), font, 0.4, txt_color, thickness=1)

    return img

# %% ../nbs/00_visualize.ipynb 15
class Board:
    """
        Example:
        board = Board()
        board.set_line_text(0, 'ok')
        board().show()
    """
    def __init__(self, num_lines=10, line_h=40, line_w = 300, font_scale=1, colors=None):
        colors = list(matplotlib.colors.ColorConverter.colors.keys())
        colors = np.array([matplotlib.colors.ColorConverter.to_rgb(c) for c in colors])
        colors = (colors*255).astype(int)        
        
        self.texts = ['' for _ in range(num_lines)]
        self.board_h = line_h*(num_lines+2)
        self.board_w = line_w
        self.line_h = line_h
        self.font_scale = 1
        self.colors = colors
        self.img_board = None
        if colors is None:
            self.colors = mplCOCO_COLORS[:num_lines]
    
    def draw(self):
        for i, text in enumerate(self.texts):
            color = self.colors[i].tolist()
            if isinstance(text, tuple):
                text, prob = text
                start_x = len(text)*23
                p1 = start_x, i*self.line_h+10
                p2 = start_x+int(prob * 100), (i+1)*self.line_h
                cv2.rectangle(self.board, p1, p2, color, -1)
                
                p1 = start_x, i*self.line_h+10
                p2 = start_x+100, (i+1)*self.line_h
                
                cv2.rectangle(self.board, p1, p2, (255,255,255), 1)
                
                cv2.putText(self.board, text, (0, p2[1]), cv2.FONT_HERSHEY_COMPLEX, 1, color)
                
            elif len(text):
                x = 0
                y = (i+1)*self.line_h
                cv2.putText(self.board, text, (x, y), cv2.FONT_HERSHEY_COMPLEX, 1, color)
        return self.board
    
    def set_line_text(self, line_idx, text, prob=None):
        if line_idx is None:
            if not hasattr(self, 'line_idx'):
                self.line_idx = 0
            else:
                self.line_idx += 1
            line_idx = self.line_idx
        if prob is not None:
            assert prob >=0 and prob <=1
            text = (text, prob)
        self.texts[line_idx] = text
    
    def clear(self):
        self.board = np.zeros([self.board_h, self.board_w, 3], 'uint8')
    
    def __call__(self):
        self.clear()
        self.draw()
        return self
    
    def show(self):
        show(self.board)
        

    def lazy_img_concat(self, img):
        # call the returned fuction to get the actual img_concat value
        self.img = img
        def f():
            self()
            board = self.board
            img = mmcv.imread(self.img)
            board = mmcv.imresize_like(board, img)
            return np.concatenate([img, board], 1)
        return f
    
    def img_concat(self, img):
        return self.lazy_img_concat(img)()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_debug.ipynb.

# %% auto 0
__all__ = ['make_mini_coco', 'dpython']

# %% ../nbs/04_debug.ipynb 2
from ._imports import *

# %% ../nbs/04_debug.ipynb 3
@call_parse
def make_mini_coco(json_path: Param(),
                   image_prefix: Param(),
                   out_dir: Param(),
                   num_samples: Param("Num of sample",type=int) = 1000):
    """
        Helper function for creating a mini-dataset ensembles it's father
    """
    new_img_prefix = osp.join(out_dir, "images")

    out_json = os.path.join(out_dir, "annotations", "mini_json.json")
    if not osp.exists(out_json):
        logger.info(f"Making mini dataset out_dir-> {out_dir}, num images:{num_samples}")
        os.makedirs(os.path.join(out_dir, "images"), exist_ok=True)
        os.makedirs(os.path.join(out_dir, "annotations"), exist_ok=True)
        coco = COCO(json_path)
        # imgs = coco.imgs
        img_ids = list(coco.imgs.keys())
        np.random.seed(0)
        selected_img_ids = np.random.choice(img_ids, num_samples, replace=False)
        imgs = coco.loadImgs(selected_img_ids)
        selected_ann_ids = coco.getAnnIds(selected_img_ids)
        anns = coco.loadAnns(selected_ann_ids)
        for i, ann in enumerate(anns):
            ann['iscrowd'] = False
            anns[i] = ann
        out_dict = dict(
            images=imgs,
            annotations=anns,
            categories=coco.dataset['categories'],

        )
        for img in imgs:
            path = osp.join(image_prefix, img['file_name'])
            new_path = osp.join(new_img_prefix, img['file_name'])
            mmcv.mkdir_or_exist(osp.dirname(new_path))
            shutil.copy(path, new_path)

        mmcv.dump(out_dict, out_json)
    logger.info(f"{out_json}, {new_img_prefix}")
    return out_json, new_img_prefix

# %% ../nbs/04_debug.ipynb 4
@call_parse
def dpython(cmd: Param(type=str)):
    for _ in range(3):
        cmd = cmd.replace('  ', '')
    i_split = cmd.index(".py")+4
    file = cmd[:i_split].strip().split(' ')[1]

    args = cmd[i_split:].split(' ')
    cfg_name = os.environ.get("DNAME", "Latest-Generated")
    cfg = {
        "name": f"Python: {cfg_name}",
        "type": "python",
        "request": "launch",
        "program": f"{file}",
        "console": "integratedTerminal",
        "args": args,
    }
    # pp(cfg)
    mmcv.mkdir_or_exist(".vscode")
    try:
        lauch = mmcv.load(".vscode/launch.json")
    except Exception as e:
        lauch = {
            "version": "0.2.0",
            "configurations": [

            ]
        }
        logger.warning(e)
        
    replace = False
    for i, _cfg in enumerate(lauch['configurations']):
        if _cfg["name"] == cfg["name"]:
            lauch["configurations"][i] = cfg
            replace = True
    if not replace:
        lauch["configurations"] += [cfg]
        mmcv.dump(lauch, '.vscode/launch.json')



from avcv._imports import *
from avcv.utils import *
from avcv.process import *
from avcv.visualize import *
from avcv.coco import *

# import cv2, matplotlib.pyplot as plt, mmcv, numpy as np, os, os.path as osp
# from glob import glob
# from tqdm import tqdm
# import pandas as pd
# from PIL import Image
# from loguru import logger


# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_process.ipynb.

# %% auto 0
__all__ = ['multi_thread', 'multi_process']

# %% ../nbs/01_process.ipynb 2
from ._imports import *

# %% ../nbs/01_process.ipynb 3
def multi_thread(fn, array_inputs, max_workers=None, 
                 desc="", unit="Samples", 
                 verbose=True, pbar_iterval=10):

    def _wraper(x):
        i, input = x
        return {i: fn(input)}

    array_inputs = [(i, _) for i, _ in enumerate(array_inputs)]
    if verbose:
        logger.info(desc)
        progress_bar = mmcv.ProgressBar(len(array_inputs))#tqdm(total=len(array_inputs))
    outputs = {}
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        for i, result in enumerate(executor.map(_wraper, array_inputs)):
            outputs.update(result)
            if verbose and i%pbar_iterval==0:
                progress_bar.update(pbar_iterval)
    # if verbose:
        # logger.info('multi_thread {}, {}', fn.__name__, desc)
    outputs = list(outputs.values())
    return outputs




# %% ../nbs/01_process.ipynb 5
def multi_process(f, inputs, max_workers=8, desc='',
               unit='Samples', verbose=True, pbar_iterval=10):
    if verbose:
        
        logger.info('Multi processing {} | Num samples: {}', f.__name__, len(inputs))
        pbar = mmcv.ProgressBar(len(inputs))
        
    with Pool(max_workers) as p:
        it = p.imap(f, inputs)
        total = len(inputs)
        # return list(tqdm(it, total=total))
        return_list = []
        for i, ret in enumerate(it):
            return_list.append(ret)
            if i % pbar_iterval == 0 and verbose:
                pbar.update(pbar_iterval)
    return return_list





# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/avcv/',
                'doc_host': 'https://anhvth.github.io',
                'git_url': 'https://github.com/anhvth/avcv/tree/nbdev-convert/',
                'lib_path': 'avcv'},
  'syms': { 'avcv.all': {},
            'avcv.cli': {'avcv.cli.convert_image': ('cli.html#convert_image', 'avcv/cli.py')},
            'avcv.coco': { 'avcv.coco.AvCOCO': ('coco_dataset.html#avcoco', 'avcv/coco.py'),
                           'avcv.coco.AvCOCO.__init__': ('coco_dataset.html#avcoco.__init__', 'avcv/coco.py'),
                           'avcv.coco.AvCOCO.createIndex': ('coco_dataset.html#avcoco.createindex', 'avcv/coco.py'),
                           'avcv.coco.AvCOCO.loadRes': ('coco_dataset.html#avcoco.loadres', 'avcv/coco.py'),
                           'avcv.coco.CocoDataset': ('coco_dataset.html#cocodataset', 'avcv/coco.py'),
                           'avcv.coco.CocoDataset.__init__': ('coco_dataset.html#cocodataset.__init__', 'avcv/coco.py'),
                           'avcv.coco.CocoDataset.evaluate': ('coco_dataset.html#cocodataset.evaluate', 'avcv/coco.py'),
                           'avcv.coco.CocoDataset.imread': ('coco_dataset.html#cocodataset.imread', 'avcv/coco.py'),
                           'avcv.coco.CocoDataset.load_anns': ('coco_dataset.html#cocodataset.load_anns', 'avcv/coco.py'),
                           'avcv.coco.CocoDataset.visualize': ('coco_dataset.html#cocodataset.visualize', 'avcv/coco.py'),
                           'avcv.coco.DiagnoseCoco': ('coco_dataset.html#diagnosecoco', 'avcv/coco.py'),
                           'avcv.coco.DiagnoseCoco.find_false_samples': ( 'coco_dataset.html#diagnosecoco.find_false_samples',
                                                                          'avcv/coco.py'),
                           'avcv.coco._f': ('coco_dataset.html#_f', 'avcv/coco.py'),
                           'avcv.coco.bbox_expand': ('coco_dataset.html#bbox_expand', 'avcv/coco.py'),
                           'avcv.coco.check_save_coco_dict': ('coco_dataset.html#check_save_coco_dict', 'avcv/coco.py'),
                           'avcv.coco.concat_coco': ('coco_dataset.html#concat_coco', 'avcv/coco.py'),
                           'avcv.coco.extract_coco': ('coco_dataset.html#extract_coco', 'avcv/coco.py'),
                           'avcv.coco.get_bboxes': ('coco_dataset.html#get_bboxes', 'avcv/coco.py'),
                           'avcv.coco.get_overlap_rate': ('coco_dataset.html#get_overlap_rate', 'avcv/coco.py'),
                           'avcv.coco.split_coco': ('coco_dataset.html#split_coco', 'avcv/coco.py'),
                           'avcv.coco.to_jpg': ('coco_dataset.html#to_jpg', 'avcv/coco.py'),
                           'avcv.coco.v2c': ('coco_dataset.html#v2c', 'avcv/coco.py'),
                           'avcv.coco.video_to_coco': ('coco_dataset.html#video_to_coco', 'avcv/coco.py')},
            'avcv.debug': { 'avcv.debug.dpython': ('debug.html#dpython', 'avcv/debug.py'),
                            'avcv.debug.make_mini_coco': ('debug.html#make_mini_coco', 'avcv/debug.py')},
            'avcv.dist_utils': { 'avcv.dist_utils.all_gather': ('dist_utils.html#all_gather', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.all_gatherv': ('dist_utils.html#all_gatherv', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.all_reduce': ('dist_utils.html#all_reduce', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.gather_grad': ('dist_utils.html#gather_grad', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.get_rank': ('dist_utils.html#get_rank', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.get_world_size': ('dist_utils.html#get_world_size', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.primary': ('dist_utils.html#primary', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.rank0_to_all': ('dist_utils.html#rank0_to_all', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.reduce_loss_dict': ('dist_utils.html#reduce_loss_dict', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.reduce_sum': ('dist_utils.html#reduce_sum', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.setup_distributed': ('dist_utils.html#setup_distributed', 'avcv/dist_utils.py'),
                                 'avcv.dist_utils.synchronize': ('dist_utils.html#synchronize', 'avcv/dist_utils.py')},
            'avcv.process': { 'avcv.process.multi_process': ('process.html#multi_process', 'avcv/process.py'),
                              'avcv.process.multi_thread': ('process.html#multi_thread', 'avcv/process.py')},
            'avcv.utils': { 'avcv.utils.TimeLoger': ('utils.html#timeloger', 'avcv/utils.py'),
                            'avcv.utils.TimeLoger.__init__': ('utils.html#timeloger.__init__', 'avcv/utils.py'),
                            'avcv.utils.TimeLoger.__str__': ('utils.html#timeloger.__str__', 'avcv/utils.py'),
                            'avcv.utils.TimeLoger.start': ('utils.html#timeloger.start', 'avcv/utils.py'),
                            'avcv.utils.TimeLoger.update': ('utils.html#timeloger.update', 'avcv/utils.py'),
                            'avcv.utils.VideoReader': ('utils.html#videoreader', 'avcv/utils.py'),
                            'avcv.utils.VideoReader.__getitem__': ('utils.html#videoreader.__getitem__', 'avcv/utils.py'),
                            'avcv.utils.VideoReader.__init__': ('utils.html#videoreader.__init__', 'avcv/utils.py'),
                            'avcv.utils.VideoReader.__len__': ('utils.html#videoreader.__len__', 'avcv/utils.py'),
                            'avcv.utils.av_i2v': ('utils.html#av_i2v', 'avcv/utils.py'),
                            'avcv.utils.find_contours': ('utils.html#find_contours', 'avcv/utils.py'),
                            'avcv.utils.generate_tmp_filename': ('utils.html#generate_tmp_filename', 'avcv/utils.py'),
                            'avcv.utils.get_files': ('utils.html#get_files', 'avcv/utils.py'),
                            'avcv.utils.get_md5': ('utils.html#get_md5', 'avcv/utils.py'),
                            'avcv.utils.get_name': ('utils.html#get_name', 'avcv/utils.py'),
                            'avcv.utils.identify': ('utils.html#identify', 'avcv/utils.py'),
                            'avcv.utils.images_to_video': ('utils.html#images_to_video', 'avcv/utils.py'),
                            'avcv.utils.imemoize': ('utils.html#imemoize', 'avcv/utils.py'),
                            'avcv.utils.is_interactive': ('utils.html#is_interactive', 'avcv/utils.py'),
                            'avcv.utils.md5_from_str': ('utils.html#md5_from_str', 'avcv/utils.py'),
                            'avcv.utils.memoize': ('utils.html#memoize', 'avcv/utils.py'),
                            'avcv.utils.mkdir': ('utils.html#mkdir', 'avcv/utils.py'),
                            'avcv.utils.printc': ('utils.html#printc', 'avcv/utils.py'),
                            'avcv.utils.put_text': ('utils.html#put_text', 'avcv/utils.py'),
                            'avcv.utils.self_memoize': ('utils.html#self_memoize', 'avcv/utils.py'),
                            'avcv.utils.v2i': ('utils.html#v2i', 'avcv/utils.py'),
                            'avcv.utils.video_to_images': ('utils.html#video_to_images', 'avcv/utils.py')},
            'avcv.visualize': { 'avcv.visualize.Board': ('visualize.html#board', 'avcv/visualize.py'),
                                'avcv.visualize.Board.__call__': ('visualize.html#board.__call__', 'avcv/visualize.py'),
                                'avcv.visualize.Board.__init__': ('visualize.html#board.__init__', 'avcv/visualize.py'),
                                'avcv.visualize.Board.clear': ('visualize.html#board.clear', 'avcv/visualize.py'),
                                'avcv.visualize.Board.draw': ('visualize.html#board.draw', 'avcv/visualize.py'),
                                'avcv.visualize.Board.img_concat': ('visualize.html#board.img_concat', 'avcv/visualize.py'),
                                'avcv.visualize.Board.lazy_img_concat': ('visualize.html#board.lazy_img_concat', 'avcv/visualize.py'),
                                'avcv.visualize.Board.set_line_text': ('visualize.html#board.set_line_text', 'avcv/visualize.py'),
                                'avcv.visualize.Board.show': ('visualize.html#board.show', 'avcv/visualize.py'),
                                'avcv.visualize.bbox_visualize': ('visualize.html#bbox_visualize', 'avcv/visualize.py'),
                                'avcv.visualize.imshow': ('visualize.html#imshow', 'avcv/visualize.py'),
                                'avcv.visualize.plot_images': ('visualize.html#plot_images', 'avcv/visualize.py'),
                                'avcv.visualize.tensor2imgs': ('visualize.html#tensor2imgs', 'avcv/visualize.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_cli.ipynb.

# %% auto 0
__all__ = ['convert_image']

# %% ../nbs/06_cli.ipynb 4
from glob import glob
from .utils import multi_thread
import os.path as osp
from fastcore.script import *
import mmcv
import os
from loguru import logger


@call_parse
def convert_image(path: Param(help="Root dir", type=str, default='**/*.png'),
                  type_from_to: Param(help='Convert image from one type to another default=png->jpg', type=str, default='png->jpg'),
                  remove: Param(help="Remove after converted", type=bool),
                  recursive: Param(help="glob recursive", type=bool)
                 ):
    assert '->' in 'type_from_to must include "->" '
    source_ext, target_ext = type_from_to.split('->')
    paths = glob(path, recursive=recursive)
    if remove:
        logger.info('Files will be removed after converted')
    def f_convert(path):
        dir_name, file_name = osp.dirname(path), osp.basename(path)
        new_file_name = file_name.replace(f'.{source_ext}', f'.{target_ext}')
        new_path = osp.join(dir_name, new_file_name)
        mmcv.imwrite(mmcv.imread(path), new_path)
        if remove: 
            os.remove(path)            
    multi_thread(f_convert, paths, desc=f"Converting {type_from_to} in {path}")

    
